from math import *
from adap_gl import AdapGL

def normal_0_to_z( z ):
    """
	Kerridge and Cook(1976, Biometrika, 63, 401-403)'s Algorithm

    """    
    p = None
    if z < 0.0:
        raise Exception("Negative z value in normal_0_to_z")
    else:
        if z == 0.0:
            p = 0.0
        else:
            if z > 6.8:
                zz = z * z;
                d = zz + 3 - 1 / (0.22 * zz + 0.704);
                m = 1 - 1 / d;
                p = 0.5 - exp(-0.5 * zz) * m / (z * ((2 * pi) ** 0.5));
            else:
                p = 0.0
                zz4 = 0.25 * z * z;
                n2 = 2.0;               # initial value of 2n          ; n = 1
                theta2 = 1.0;	    # initial value of theta(2n-2)
                theta1 = zz4;
                p = 1.0;		    # sum for n = 0
                p_prev = 0.0;	    # initial value of sub-sum
                while True:
                    theta2 = zz4 * (theta1 - theta2) / n2;
                    n2 += 1.0;
                    p += theta2 / n2;
                    if (p == p_prev):
                        break;
                    theta1 = zz4 * (theta2 - theta1) / n2;
                    n2 += 1.0;
                    p_prev = p;
                    
                p = z * exp(-0.5 * zz4) * p / ((2.0 * pi) ** 0.5);

    return p

def Cum_NormalKC( z ):
    """
            Cumulative normal distribution with mean 0 and sd 1
    """
    v = None
    if (z > 0.0):
        v = 0.5 + normal_0_to_z(z);
        if (v > 1.0):
            v = 1.0
    elif z == 0.0:
        v = 0.5
    else:
        v = 0.5 - normal_0_to_z(abs(z));
        if (v < 0.0):
            v = 0.0

    return v

def phi( x ):
    """
            The density function of the standard normal distribution
    """
    return exp(-0.5 * x * x) / ((2.0 * pi) ** 0.5);

"""
            In the following, Equation Numbers are those in Okamoto(2017)
"""

class CFuncCovFY_trans:
    """
            Integrand in Equation (A.9)
            with variable transformation
                    Fi = Log( (1 + t) / (1 - t) )
            Hence, we have
            Functor( t ) = Fi * {Sigma[j: 1..m]Sigma[k: 1..K]{k * P(Yij = k | Fi)}}
                                * phi(Fi) * 2.0 / (1 - t * t)
    """
    def __init__(self, K, m, mu, lmbd, psy, C):
        self.K = K
        self.m = m
        self.mu = []
        self.lmbd = []
        self.psy = []
        for i in range(m):
            self.mu.append(mu[i])
            self.lmbd.append(lmbd[i])
            self.psy.append(psy[i])

        self.C = []
        for j in range(K - 1):
            self.C.append(C[j])

            
    def v(self, t):
        Fi = log((1.0 + t) / (1.0 - t))
        v = 0.0
        for j in range(1, self.m + 1):
            for k in range(1, self.K + 1):
                tv = None
                if (k == 1):
                    tv = Cum_NormalKC((self.C[0] - (self.mu[j - 1] + self.lmbd[j - 1] * Fi)) / self.psy[j - 1])
                elif (k == self.K):
                    tv = 1.0 - Cum_NormalKC((self.C[self.K - 2] - (self.mu[j - 1] + self.lmbd[j - 1] * Fi)) / self.psy[j - 1])
                else:
                    tv = Cum_NormalKC((self.C[k - 1] - (self.mu[j - 1] + self.lmbd[j - 1] * Fi)) / self.psy[j - 1]) \
                            - Cum_NormalKC((self.C[k - 2] - (self.mu[j - 1] + self.lmbd[j - 1] * Fi)) / self.psy[j - 1])
                v += k * tv

        v = Fi * v * phi(Fi) * 2.0 / (1.0 - t * t)
        return v


class CFuncPY:
    """
            The integrand in Equation (A.4)
            with variable transformaton
                    Fi = Log( (1 + t) / (1 - t) )
            Hence, we have
            	Functor( t ) = P(Yij = k | Fi) * phi(Fi) * 2 / (1 - t * t)
    """
    def __init__(self, K, m, mu, lmbd, psy, C):
        self.k = None
        self.j = None
        self.K = K
        self.m = m
        self.mu = []
        self.lmbd = []
        self.psy = []
        for i in range(m):
            self.mu.append(mu[i])
            self.lmbd.append(lmbd[i])
            self.psy.append(psy[i])

        self.C = []
        for jj in range(K - 1):
            self.C.append(C[jj])

    def v(self, t):
        Fi = log((1.0 + t) / (1.0 - t))
        v = None
        if (self.k == 1):
            v = Cum_NormalKC((self.C[0] - (self.mu[self.j - 1] + self.lmbd[self.j - 1] * Fi)) / self.psy[self.j - 1])
        elif (self.k == self.K):
            v = 1.0 - Cum_NormalKC((self.C[self.K - 2] - (self.mu[self.j - 1] + self.lmbd[self.j - 1] * Fi)) / self.psy[self.j - 1])
        else:
            v = Cum_NormalKC((self.C[self.k - 1] - (self.mu[self.j - 1] + self.lmbd[self.j - 1] * Fi)) / self.psy[self.j - 1]) \
		- Cum_NormalKC((self.C[self.k - 2] - (self.mu[self.j - 1] + self.lmbd[self.j - 1] * Fi)) / self.psy[self.j - 1])
        
        if (v < 0.0):
            v = 0.0
        if (v > 1.0):
            v = 1.0
        v = v * phi(Fi) * 2.0 / (1.0 - t * t)

        #   print('k = {0}  j = {1}  v = {2}'.format(self.k, self.j, v))
        
        return v



#	double ** PY;


class CFuncPYY:
    """
            The integrand in Equaiton (A.8)
            with variable transfomation
                Fi = Log( (1 + t) / (1 - t) )
            Hence, we have
                Functor( t ) = P(Yij1 = k1 | Fi) * P(Yij2 = k2 | Fi) * phi(Fi) * 2 / (1 - t * t)
    """
    def __init__(self, K, m, mu, lmbd, psy, C):
        self.k1 = None
        self.k2 = None
        self.j1 = None
        self.j2 = None
        self.K = K
        self.m = m
        self.mu = []
        self.lmbd = []
        self.psy = []
        for i in range(m):
            self.mu.append(mu[i])
            self.lmbd.append(lmbd[i])
            self.psy.append(psy[i])

        self.C = []
        for j in range(K - 1):
            self.C.append(C[j])

    def v(self, t):
        Fi = log((1.0 + t) / (1.0 - t))
        v1 = None
        if (self.k1 == 1):
            v1 = Cum_NormalKC((self.C[0] - (self.mu[self.j1 - 1] + self.lmbd[self.j1 - 1] * Fi)) / self.psy[self.j1 - 1])
        elif (self.k1 == self.K):
            v1 = 1.0 - Cum_NormalKC((self.C[self.K - 2] - (self.mu[self.j1 - 1] + self.lmbd[self.j1 - 1] * Fi)) / self.psy[self.j1 - 1])
        else:
            v1 = Cum_NormalKC((self.C[self.k1 - 1] - (self.mu[self.j1 - 1] + self.lmbd[self.j1 - 1] * Fi)) / self.psy[self.j1 - 1])  \
                    - Cum_NormalKC((self.C[self.k1 - 2] - (self.mu[self.j1 - 1] + self.lmbd[self.j1 - 1] * Fi)) / self.psy[self.j1 - 1])
        if (v1 < 0.0):
            v1 = 0.0
        if (v1 > 1.0):
            v1 = 1.0

        v2 = None
        
        if (self.k2 == 1):
            v2 = Cum_NormalKC((self.C[0] - (self.mu[self.j2 - 1] + self.lmbd[self.j2 - 1] * Fi)) / self.psy[self.j2 - 1])
        elif (self.k2 == self.K):
            v2 = 1.0 - Cum_NormalKC((self.C[self.K - 2] - (self.mu[self.j2 - 1] + self.lmbd[self.j2 - 1] * Fi)) / self.psy[self.j2 - 1])
        else:
            v2 = Cum_NormalKC((self.C[self.k2 - 1] - (self.mu[self.j2 - 1] + self.lmbd[self.j2 - 1] * Fi)) / self.psy[self.j2 - 1])  \
                    - Cum_NormalKC((self.C[self.k2 - 2] - (self.mu[self.j2 - 1] + self.lmbd[self.j2 - 1] * Fi)) / self.psy[self.j2 - 1]);
        if (v2 < 0.0):
            v2 = 0.0
        if (v2 > 1.0):
            v2 = 1.0
        v = v1 * v2 * phi(Fi) * 2.0 / (1.0 - t * t)
        return v

    

#	double ** PYY;


class CalcRhoPreRhoYYP:
    """
            Calculation of reliability coefficients for
            categorical items
    """
    def __init__(self, K, m, mu, lmbd, psy, C):
        #   print('Initializing a CalcRhoPreRhoYYP object.') 
        self.K = K
        self.m = m
        self.mu = []
        self.lmbd = []
        self.psy = []
        for i in range(m):
            self.mu.append(mu[i])
            self.lmbd.append(lmbd[i])
            self.psy.append(psy[i])

        self.C = []
        for j in range(K - 1):
            self.C.append(C[j])

        #   print('Initialization ended.')

        """
                Coavariacce between Fi and Yi
        """
        fnctr_trans = CFuncCovFY_trans (K, m, mu, lmbd, psy, C)
        a_gl = AdapGL()
        cov = a_gl.integral(-1.0, 1.0, fnctr_trans.v, 1.0e-9)

        #   print('cov was calculated...{}'.format(cov))

        """
                PY[j-1][k-1] = P(Yij = k)
        """

        fnctr_PY = CFuncPY(K, m, mu, lmbd, psy, C)

        PY = []
        for j in range(1, m + 1):
            PY.append([])
            for k in range(1, K + 1):
                fnctr_PY.k = k
                fnctr_PY.j = j
                v = a_gl.integral(-1.0, 1.0, fnctr_PY.v, 1.0e-9)
                PY[j - 1].append(v)

                
        """
                    EY = E(Yi)

        """

        EY = 0.0
        for j in range(1, m + 1):
            for k in range(1, K + 1):
                EY += k * PY[j - 1][k - 1]

        """
                EYj2 = E(Yij**2)

        """
        EYj2 = []
        for j in range(1, m + 1):
            v = 0.0
            for k in range(1, K + 1):
                v += k * k * PY[j - 1][k - 1]
            EYj2.append(v)

        #   print('EYj2 = ')
        #   print(EYj2)
        

        """
                PYY[j1-1][k1-1][j2-1][k2-1] = P(Yij1 = k1, Yij2 = k2)

        """
        PYY = []
        for j1 in range(1, m + 1):
            PYY.append([])
            for k1 in range(1, K + 1):
                PYY[j1 - 1].append([])
                for j2 in range(1, m + 1):
                    PYY[j1 - 1][k1 - 1].append([])
                    for k2 in range(1, K + 1):
                        fnctr_PYY = CFuncPYY(K, m, mu, lmbd, psy, C)

                        fnctr_PYY.j1 = j1
                        fnctr_PYY.k1 = k1
                        fnctr_PYY.j2 = j2
                        fnctr_PYY.k2 = k2

                        v = a_gl.integral(-1.0, 1.0, fnctr_PYY.v, 1.0e-9)
                        PYY[j1 - 1][k1 - 1][j2 - 1].append(v)
                        #   print('v = {}'.format(v))


        """
                EYY[j1-1][j2-1] = E(Yij1 * Yij2)

        """
        EYY = []

        for j1 in range(1, m + 1):
            EYY.append([])
            for j2 in range(1, m + 1):
                v = 0.0
                for k1 in range(1, K + 1):
                    for k2 in range(1, K + 1):
                        v += k1 * k2 * PYY[j1 - 1][k1 - 1][j2 - 1][k2 - 1]

                EYY[j1 - 1].append(v)

        """
                    EY2 = E(Yi**2)      Equation (A.5)

        """
        EY2 = 0.0
        for j in range(m):
            EY2 += EYj2[j]

        for j1 in range(m):
            for j2 in range(m):
                if (j1 != j2):
                    EY2 += EYY[j1][j2]

        """
                    Var_Yi = Var(Yi) 
        """
        Var_Yi = EY2 - (EY ** 2);

        #   print('Var_Yi = {0}   EY2 = {1}   EY = {2}'.format(Var_Yi, EY2, EY))


        """
                RhoPre = Rho_pre:  Predictive reliability coefficient

        """
        self.RhoPre = (cov ** 2) / Var_Yi
        
        EYYP = 0.0
        for j1 in range(m):
            for j2 in range(m):
                EYYP += EYY[j1][j2]

        CovYYP = EYYP - (EY ** 2)


        """
                    RhoYYP = Cor(Y, Y'),   Y' parallel to Y

        """
        self.RhoYYP = CovYYP / Var_Yi;


        """
                    EYj[j] = E(Yij)       Equation (A.16)

        """
        EYj = []
        for j in range(m):
            v = 0.0
            for k in range(1, K + 1):
                v += k * PY[j][k - 1]
            EYj.append(v)

        """
                    sjj[j] = Var(Yij)     Equation (A.15)

        """

        sjj = []
        for j in range(m):
            v = EYj2[j] - ((EYj[j]) ** 2)
            sjj.append(v)

        sum_sjj = 0.0
        for j in range(m):
            sum_sjj += sjj[j];

        """
                coeff_a = coefficient alpha,   Equation (A.14)

        """
        self.coeff_a = m * (1.0 - sum_sjj / Var_Yi) / (m - 1.0)
            

        """
                    rho_omega = coefficient omega     Equation (6)

        """

        num_rho = 0.0
        sum_psy = 0.0
        for j in range(m):
            num_rho += self.lmbd[j]
            sum_psy += (self.psy[j]) ** 2

        self.rho_omega = (num_rho ** 2) / ((num_rho ** 2) + sum_psy);


    def GetRhoPre(self):
        return self.RhoPre


    def GetRhoYYP(self):
        
        return self.RhoYYP


    def GetCoeffAlpha(self):
        return self.coeff_a


    def GetOmega(self):
        return self.rho_omega




